#include <stdlib.h>

#include <X11/XCB/xcb.h>

typedef XCBVoidCookie Ecore_Xcb_Generic_Cookie;
typedef XCBConnection Ecore_Xcb_Connection;
typedef XCBGenericError Ecore_Xcb_Generic_Error;

extern XCBConnection *ecore_xcb_conn;

typedef void *(*Ecore_Xcb_Reply_Func) (Ecore_Xcb_Connection      *c,
                                       Ecore_Xcb_Generic_Cookie   cookie,
                                       Ecore_Xcb_Generic_Error  **e);

typedef struct _Ecore_Xcb_Request Ecore_Xcb_Request;

struct _Ecore_Xcb_Request
{
  Ecore_Xcb_Generic_Cookie cookie;
  Ecore_Xcb_Reply_Func     reply_function;
  void                    *reply;
};

static int                ecore_xcb_request_init_count = 0;
static int                ecore_xcb_request_ref        = 0;
static int                ecore_xcb_request_count      = 16; /* Number of requests during ecore init: 140 */
static Ecore_Xcb_Request *ecore_xcb_request_requests   = NULL;

static void ecore_xcb_request_resize (void);

void
ecore_xcb_request_init (int request_count)
{
  if (!ecore_xcb_request_init_count)
    {
      if (request_count > 0)
        ecore_xcb_request_count = request_count;

      ecore_xcb_request_requests = (Ecore_Xcb_Request *)malloc (ecore_xcb_request_count * sizeof (Ecore_Xcb_Request));
      if (!ecore_xcb_request_requests)
        return;
    }
  ecore_xcb_request_init_count++;
}

int
ecore_xcb_request_add (Ecore_Xcb_Generic_Cookie *cookie,
                       Ecore_Xcb_Reply_Func      reply_function)
{
  unsigned int id = -1;

  if (!cookie || !reply_function)
    return id;

  if (ecore_xcb_request_ref + 1 == ecore_xcb_request_count)
    ecore_xcb_request_resize ();

  ecore_xcb_request_requests[ecore_xcb_request_ref].cookie = *cookie;
  ecore_xcb_request_requests[ecore_xcb_request_ref].reply_function = reply_function;
  ecore_xcb_request_requests[ecore_xcb_request_ref].reply = NULL;
  id = ecore_xcb_request_ref;

  ecore_xcb_request_ref++;

  return id;
}

void *
ecore_xcb_request_poll (int id)
{
  if ((id >= 0) &&
      (id < ecore_xcb_request_ref) &&
      !ecore_xcb_request_requests[id].reply)
    {
      ecore_xcb_request_requests[id].reply = ecore_xcb_request_requests[id].reply_function (ecore_xcb_conn,
                                                                                            ecore_xcb_request_requests[id].cookie,
                                                                                            NULL);
      
    }
  return ecore_xcb_request_requests[id].reply;
}

void
ecore_xcb_request_flush ()
{
  int rep = 0;

  while (rep != ecore_xcb_request_ref)
    {
      if (!ecore_xcb_request_requests[rep].reply)
        ecore_xcb_request_requests[rep].reply =
          ecore_xcb_request_requests[rep].reply_function (ecore_xcb_conn,
                                                          ecore_xcb_request_requests[rep].cookie,
                                                          NULL);
      
      free (ecore_xcb_request_requests[rep].reply);
      rep++;
    }
  ecore_xcb_request_ref = 0;
  ecore_xcb_request_count = 16;
  ecore_xcb_request_requests = realloc (ecore_xcb_request_requests, ecore_xcb_request_count * sizeof (Ecore_Xcb_Request));
}

void
ecore_xcb_request_shutdown ()
{
  ecore_xcb_request_init_count--;
  if (ecore_xcb_request_init_count > 0)
    return;

  ecore_xcb_request_flush ();

  ecore_xcb_request_ref = 0;

  ecore_xcb_request_count = 16;

  if (ecore_xcb_request_requests)
    free (ecore_xcb_request_requests);
  ecore_xcb_request_requests = NULL;

  ecore_xcb_request_init_count = 0;
}

/* FIXME: should we test if realloc fails ? */
static void
ecore_xcb_request_resize (void)
{
  ecore_xcb_request_count += 16;

  ecore_xcb_request_requests = realloc (ecore_xcb_request_requests, ecore_xcb_request_count * sizeof (Ecore_Xcb_Request));
}
